import io
import numbers
import warnings
from typing import List, Sequence, Tuple, Union

import numpy as np

import tiledb.libtiledb as lt

from .attribute import Attr
from .ctx import Ctx, CtxMixin, default_ctx
from .dimension_label import DimLabel
from .domain import Domain
from .filter import Filter, FilterList

if lt.version()[0] == 2 and lt.version()[1] >= 25:
    from .current_domain import CurrentDomain

_tiledb_order_to_string = {
    lt.LayoutType.ROW_MAJOR: "row-major",
    lt.LayoutType.COL_MAJOR: "col-major",
    lt.LayoutType.GLOBAL_ORDER: "global",
    lt.LayoutType.UNORDERED: "unordered",
    lt.LayoutType.HILBERT: "hilbert",
}

_string_to_tiledb_order = {v: k for k, v in _tiledb_order_to_string.items()}
_string_to_tiledb_order.update(
    {
        "C": lt.LayoutType.ROW_MAJOR,
        "R": lt.LayoutType.COL_MAJOR,
        "H": lt.LayoutType.HILBERT,
        "U": lt.LayoutType.UNORDERED,
        None: lt.LayoutType.ROW_MAJOR,  # default (fixed in SC-27374)
    }
)


class ArraySchema(CtxMixin, lt.ArraySchema):
    """
    Schema class for TileDB dense / sparse array representations

    :param domain: Domain of schema
    :param attrs: tuple of attributes
    :param cell_order:  TileDB label for cell layout
    :param tile_order:  TileDB label for tile layout
    :param int capacity: tile cell capacity
    :param offsets_filters: (default None) offsets filter list
    :param validity_filters: (default None) validity filter list
    :param bool allows_duplicates: True if duplicates are allowed
    :param bool sparse: True if schema is sparse, else False \
        (set by SparseArray and DenseArray derived classes)
    :param dim_labels: dict(dim_index, dict(dim_name, tiledb.DimSchema))
    :param enums: list of enumeration names
    :param tiledb.Ctx ctx: A TileDB Context
    :raises: :py:exc:`tiledb.TileDBError`

    """

    def __init__(
        self,
        domain: Domain = None,
        attrs: Sequence[Attr] = (),
        cell_order: str = "row-major",
        tile_order: str = "row-major",
        capacity: int = 0,
        coords_filters: Union[FilterList, Sequence[Filter]] = None,
        offsets_filters: Union[FilterList, Sequence[Filter]] = None,
        validity_filters: Union[FilterList, Sequence[Filter]] = None,
        allows_duplicates: bool = False,
        sparse: bool = False,
        dim_labels={},
        enums=None,
        ctx: Ctx = None,
    ):
        super().__init__(ctx, lt.ArrayType.SPARSE if sparse else lt.ArrayType.DENSE)

        if enums is not None:
            for enum_name in enums:
                self._add_enumeration(self._ctx, enum_name)

        if attrs is not None:
            for att in attrs:
                if not isinstance(att, Attr):
                    raise TypeError(
                        "Cannot create schema with non-Attr value for 'attrs' argument"
                    )
                self._add_attr(att)

        try:
            self._cell_order = _string_to_tiledb_order.get(cell_order)
        except (TypeError, ValueError):
            raise ValueError(f"unknown tiledb layout: {cell_order}")

        try:
            self._tile_order = _string_to_tiledb_order.get(tile_order)
        except (TypeError, ValueError):
            raise ValueError(f"unknown tiledb layout: {tile_order}")

        if capacity > 0:
            self._capacity = capacity

        if coords_filters is not None:
            warnings.warn(
                "coords_filters is deprecated; set the FilterList for each dimension",
                DeprecationWarning,
            )

            self._coords_filters = FilterList()

            dims_with_coords_filters = []
            for dim in domain:
                dim._filters = FilterList(coords_filters)
                dims_with_coords_filters.append(dim)
            domain = Domain(dims_with_coords_filters)

        if domain is not None:
            self._domain = domain

        if offsets_filters is not None:
            self._offsets_filters = FilterList(offsets_filters)

        if validity_filters is not None:
            self._validity_filters = FilterList(validity_filters)

        self._allows_dups = allows_duplicates

        for dim_index, labels_on_dim in dim_labels.items():
            for label_name, label_schema in labels_on_dim.items():
                self._add_dim_label(self._ctx, label_name, dim_index, label_schema)

        self._check()

    @classmethod
    def load(cls, uri, ctx: Ctx = None, key: str = None):
        if not ctx:
            ctx = default_ctx()

        args = [ctx, uri]
        if key is not None:
            args.extend((lt.EncryptionType.AES_256_GCM, key))

        return cls.from_pybind11(ctx, lt.ArraySchema(*args))

    @classmethod
    def from_file(cls, uri: str = None, ctx: Ctx = None):
        """Create an ArraySchema for a Filestore Array from a given file.
        If a uri is not given, then create a default schema."""
        schema = lt.Filestore._schema_create(ctx or default_ctx(), uri)
        return cls.from_pybind11(ctx, schema)

    def __eq__(self, other):
        """Instance is equal to another ArraySchema"""
        if not isinstance(other, ArraySchema):
            return False
        if not (
            self.sparse == other.sparse
            and self.cell_order == other.cell_order
            and self.tile_order == other.tile_order
            and self.capacity == other.capacity
            and self.coords_filters == other.coords_filters
            and self.offsets_filters == other.offsets_filters
            and self.validity_filters == other.validity_filters
            and self.nattr == other.nattr
            and self.domain == other.domain
        ):
            return False
        if self.sparse and self.allows_duplicates != other.allows_duplicates:
            return False
        for i in range(self.nattr):
            if self.attr(i) != other.attr(i):
                return False
        return True

    def __len__(self):
        """Returns the number of Attributes in the ArraySchema"""
        return self._nattr

    def __iter__(self):
        """Returns a generator object that iterates over the ArraySchema's Attribute objects"""
        return (self.attr(i) for i in range(self.nattr))

    @property
    def ctx(self) -> Ctx:
        """The array schema's context

        :rtype: tiledb.Ctx
        """
        return self._ctx

    def check(self) -> bool:
        """Checks the correctness of the array schema

        :rtype: None
        :raises: :py:exc:`tiledb.TileDBError` if invalid
        """
        return self._check()

    @property
    def sparse(self) -> bool:
        """True if the array is a sparse array representation

        :rtype: bool
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return self._array_type == lt.ArrayType.SPARSE

    @property
    def allows_duplicates(self) -> bool:
        """Returns True if the (sparse) array allows duplicates."""

        if not self.sparse:
            raise lt.TileDBError(
                "ArraySchema.allows_duplicates does not apply to dense arrays"
            )

        return self._allows_dups

    @property
    def capacity(self) -> int:
        """The array capacity

        :rtype: int
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return self._capacity

    @property
    def cell_order(self) -> str:
        """The cell order layout of the array.

        :rtype: str
        """
        return _tiledb_order_to_string[self._cell_order]

    @property
    def tile_order(self) -> str:
        """The tile order layout of the array.

        :rtype: str
        :raises: :py:exc:`tiledb.TileDBError`

        """
        layout_string = _tiledb_order_to_string[self._tile_order]
        return layout_string if self.cell_order != "hilbert" else None

    @property
    def offsets_filters(self) -> FilterList:
        """The FilterList for the array's variable-length attribute offsets

        :rtype: tiledb.FilterList
        :raises: :py:exc:`tiledb.TileDBError`
        """
        return FilterList.from_pybind11(self._ctx, self._offsets_filters)

    @property
    def coords_filters(self) -> FilterList:
        """The FilterList for the array's coordinates

        :rtype: tiledb.FilterList
        :raises: :py:exc:`tiledb.TileDBError`
        """
        return FilterList.from_pybind11(self._ctx, self._coords_filters)

    @coords_filters.setter
    def coords_filters(self, value):
        warnings.warn(
            "coords_filters is deprecated; set the FilterList for each dimension",
            DeprecationWarning,
        )

    @property
    def validity_filters(self) -> FilterList:
        """The FilterList for the array's validity

        :rtype: tiledb.FilterList
        :raises: :py:exc:`tiledb.TileDBError`
        """
        return FilterList.from_pybind11(self._ctx, self._validity_filters)

    @property
    def domain(self) -> Domain:
        """The Domain associated with the array.

        :rtype: tiledb.Domain
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return Domain.from_pybind11(self._ctx, self._domain)

    @property
    def nattr(self) -> int:
        """The number of array attributes.

        :rtype: int
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return self._nattr

    @property
    def ndim(self) -> int:
        """The number of array domain dimensions.

        :rtype: int
        """
        return self.domain.ndim

    @property
    def shape(self) -> Tuple[np.dtype, np.dtype]:
        """The array's shape

        :rtype: tuple(numpy scalar, numpy scalar)
        :raises TypeError: floating point (inexact) domain
        """
        return self.domain.shape

    @property
    def version(self) -> int:
        """The array's schema (storage) version.

        :rtype: int
        :raises: :py:exc:`tiledb.TileDBError`
        """
        return self._version

    def _needs_var_buffer(self, name: str) -> bool:
        """
        Returns true if the given attribute or dimension is var-sized
        :param name:
        :rtype: bool
        """
        if self.has_attr(name):
            return self.attr(name).isvar
        elif self.domain.has_dim(name):
            return self.domain.dim(name).isvar
        else:
            raise ValueError(
                f"Requested name '{name}' is not an attribute or dimension"
            )

    def attr(self, key: Union[str, int]) -> Attr:
        """Returns an Attr instance given an int index or string label

        :param key: attribute index (positional or associative)
        :type key: int or str
        :rtype: tiledb.Attr
        :return: The ArraySchema attribute at index or with the given name (label)
        :raises TypeError: invalid key type

        """
        if isinstance(key, str):
            return Attr.from_pybind11(self._ctx, self._attr(key))
        elif isinstance(key, numbers.Integral):
            return Attr.from_pybind11(self._ctx, self._attr(int(key)))
        raise TypeError(
            "attr indices must be a string name, "
            "or an integer index, not {0!r}".format(type(key))
        )

    @property
    def attr_names(self) -> List[str]:
        """Returns a list of attribute names

        :rtype: list
        """
        return [self.attr(i).name for i in range(self.nattr)]

    def dim_label(self, name: str) -> DimLabel:
        """Returns a TileDB DimensionLabel given the label name

        :param name: name of the dimensin label
        :return: The dimension label associated with the given name
        """
        return DimLabel.from_pybind11(self._ctx, self._dim_label(self._ctx, name))

    def has_attr(self, name: str) -> bool:
        """Returns true if the given name is an Attribute of the ArraySchema

        :param name: attribute name
        :rtype: boolean
        """
        return self._has_attribute(name)

    def has_dim_label(self, name: str) -> bool:
        """Returns true if the given name is a DimensionLabel of the ArraySchema

        Note: If using an version of libtiledb that does not support dimension labels
        this will return false.

        :param name: dimension label name
        :rtype: boolean
        """
        return self._has_dim_label(self._ctx, name)

    if lt.version()[0] == 2 and lt.version()[1] >= 25:

        @property
        def current_domain(self) -> CurrentDomain:
            """Get the current domain

            :rtype: tiledb.CurrentDomain
            """
            return CurrentDomain.from_pybind11(
                self._ctx, self._current_domain(self._ctx)
            )

        def set_current_domain(self, current_domain):
            """Set the current domain

            :param current_domain: The current domain to set
            :type current_domain: tiledb.CurrentDomain
            """
            self._set_current_domain(self._ctx, current_domain)

    def attr_or_dim_dtype(self, name: str) -> bool:
        if self.has_attr(name):
            dtype = self.attr(name).dtype
        elif self.domain.has_dim(name):
            dtype = self.domain.dim(name).dtype
        else:
            raise lt.TileDBError(f"Unknown attribute or dimension ('{name}')")

        if dtype.itemsize == 0:
            # special handling for flexible numpy dtypes: change itemsize from 0 to 1
            dtype = np.dtype((dtype, 1))
        return dtype

    def dump(self):
        """Dumps a string representation of the array object to standard output (stdout)"""
        print(self._dump(), "\n")

    def __repr__(self):
        # use safe repr if pybind11 constructor failed or the array schema did
        # not construct properly
        try:
            self._check()
        except lt.TileDBError:
            return object.__repr__(self)

        if self._ctx is None:
            return object.__repr__(self)

        # TODO support/use __qualname__
        output = io.StringIO()
        output.write("ArraySchema(\n")
        output.write("  domain=Domain(*[\n")
        for i in range(self.domain.ndim):
            output.write(f"    {repr(self.domain.dim(i))},\n")
        output.write("  ]),\n")
        output.write("  attrs=[\n")
        for i in range(self.nattr):
            output.write(f"    {repr(self.attr(i))},\n")
        output.write("  ],\n")
        output.write(
            f"  cell_order='{self.cell_order}',\n"
            f"  tile_order={repr(self.tile_order)},\n"
        )
        if self.sparse:
            output.write(f"  capacity={self.capacity},\n")
        output.write(f"  sparse={self.sparse},\n")
        if self.sparse:
            output.write(f"  allows_duplicates={self.allows_duplicates},\n")

        output.write(")\n")

        return output.getvalue()

    def _repr_html_(self):
        output = io.StringIO()

        output.write("<table>")

        output.write("<tr><th>Domain</th></tr>")
        output.write(f"<tr><td>{self.domain._repr_html_()}</td></tr>")

        output.write("<tr><th>Attributes</th></tr>")
        output.write("<tr>")
        output.write("<td>")
        output.write("<table>")
        output.write("<tr>")
        output.write("<th>Name</th>")
        output.write("<th>Data Type</th>")
        output.write("<th>Is Var-Len</th>")
        output.write("<th>Is Nullable</th>")
        output.write("<th>Filters</th>")
        output.write("</tr>")
        for i in range(self.nattr):
            output.write(f"{self.attr(i)._repr_html_row_only_()}")
        output.write("</table>")
        output.write("</td>")
        output.write("</tr>")

        output.write("<tr><th>Cell Order</th></tr>")
        output.write(f"<tr><td>{self.cell_order}</td></tr>")

        output.write("<tr><th>Tile Order</th></tr>")
        output.write(f"<tr><td>{self.tile_order}</td></tr>")

        if self.sparse:
            output.write("<tr><th>Capacity</th></tr>")
            output.write(f"<tr><td>{self.capacity}</td></tr>")

        output.write("<tr><th>Sparse</th></tr>")
        output.write(f"<tr><td>{self.sparse}</td></tr>")

        if self.sparse:
            output.write("<tr><th>Allows Duplicates</th></tr>")
            output.write(f"<tr><td>{self.allows_duplicates}</td></tr>")

        output.write("</table>")

        return output.getvalue()



# quickstart_sparse.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2020 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# DESCRIPTION
#
# Please refer to the TileDB and TileDB-Py documentation for more information:
#   https://docs.tiledb.com/main/how-to
#   https://tiledb-inc-tiledb.readthedocs-hosted.com/projects/tiledb-py/en/stable/python-api.html
#
# When run, this program will create a simple 2D sparse array, write some data
# to it, and read a slice of the data back.
#


import numpy as np

import tiledb

# Name of the array to create.
array_name = "quickstart_sparse"


def create_array():
    # The array will be 4x4 with dimensions "rows" and "cols", with domain [1,4].
    dom = tiledb.Domain(
        tiledb.Dim(name="rows", domain=(1, 4), tile=4, dtype=np.int32),
        tiledb.Dim(name="cols", domain=(1, 4), tile=4, dtype=np.int32),
    )

    # The array will be sparse with a single attribute "a" so each (i,j) cell can store an integer.
    schema = tiledb.ArraySchema(
        domain=dom, sparse=True, attrs=[tiledb.Attr(name="a", dtype=np.int32)]
    )

    # Create the (empty) array on disk.
    tiledb.SparseArray.create(array_name, schema)


def write_array():
    # Open the array and write to it.
    with tiledb.SparseArray(array_name, mode="w") as A:
        # Write some simple data to cells (1, 1), (2, 4) and (2, 3).
        IJ = [1, 2, 2], [1, 4, 3]
        data = np.array(([1, 2, 3]))
        A[IJ] = data


def read_array():
    # Open the array and read from it.
    with tiledb.SparseArray(array_name, mode="r") as A:
        # Slice only rows 1, 2 and cols 2, 3, 4.
        data = A[1:3, 2:5]
        a_vals = data["a"]
        for i, coord in enumerate(zip(data["rows"], data["cols"])):
            print("Cell (%d, %d) has data %d" % (coord[0], coord[1], a_vals[i]))


if tiledb.object_type(array_name) != "array":
    create_array()
    write_array()

read_array()



import itertools
import tempfile

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import tiledb
import tiledb.libtiledb as lt

from .common import DiskTestCase

pd = pytest.importorskip("pandas")

if not (lt.version()[0] == 2 and lt.version()[1] >= 25):
    pytest.skip(
        "CurrentDomain is only available in TileDB 2.26 and later",
        allow_module_level=True,
    )


class NDRectangleTest(DiskTestCase):
    def test_ndrectagle_standalone_string(self):
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="d1", dtype="S0"),
            tiledb.Dim(name="d2", dtype="S0"),
        )
        ndrect = tiledb.NDRectangle(ctx, dom)

        range_one = ("a", "c")
        range_two = ("b", "db")

        ndrect.set_range(0, range_one[0], range_one[1])
        ndrect.set_range(1, range_two[0], range_two[1])

        self.assertEqual(ndrect.range(0), range_one)
        self.assertEqual(ndrect.range(1), range_two)

        # should be the same
        self.assertEqual(ndrect.range("d1"), range_one)
        self.assertEqual(ndrect.range("d2"), range_two)

    def test_ndrectagle_standalone_integer(self):
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="x", domain=(0, 100), tile=10, dtype=np.int64),
            tiledb.Dim(name="y", domain=(0, 100), tile=10, dtype=np.int64),
        )
        ndrect = tiledb.NDRectangle(ctx, dom)

        range_one = (10, 20)
        range_two = (30, 40)

        ndrect.set_range(0, range_one[0], range_one[1])
        ndrect.set_range(1, range_two[0], range_two[1])

        self.assertEqual(ndrect.range(0), range_one)
        self.assertEqual(ndrect.range(1), range_two)

        # should be the same
        self.assertEqual(ndrect.range("x"), range_one)
        self.assertEqual(ndrect.range("y"), range_two)

    @pytest.mark.parametrize(
        "dtype_, range_",
        (
            (np.int32, (0, 1)),
            (np.int64, (2, 7)),
            (np.uint32, (1, 5)),
            (np.uint64, (5, 9)),
            (np.float32, (0.0, 1.0)),
            (np.float64, (2.0, 4.0)),
            (np.dtype(bytes), (b"abc", b"def")),
        ),
    )
    def test_set_range_different_types(self, dtype_, range_):
        domain = tiledb.Domain(
            *[
                tiledb.Dim(
                    name="rows",
                    domain=(0, 9),
                    dtype=dtype_,
                ),
            ]
        )

        ctx = tiledb.Ctx()
        ndrect = tiledb.NDRectangle(ctx, domain)
        ndrect.set_range("rows", range_[0], range_[1])


class CurrentDomainTest(DiskTestCase):
    def test_current_domain_with_ndrectangle_integer(self):
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="x", domain=(0, 100), tile=10, dtype=np.int64),
            tiledb.Dim(name="y", domain=(0, 100), tile=10, dtype=np.int64),
        )
        ndrect = tiledb.NDRectangle(ctx, dom)

        range_one = (10, 20)
        range_two = (30, 40)

        ndrect.set_range(0, range_one[0], range_one[1])
        ndrect.set_range(1, range_two[0], range_two[1])

        self.assertEqual(ndrect.range(0), range_one)
        self.assertEqual(ndrect.range(1), range_two)

        current_domain = tiledb.CurrentDomain(ctx)

        self.assertTrue(current_domain.is_empty)
        current_domain.set_ndrectangle(ndrect)
        self.assertFalse(current_domain.is_empty)

        # let's try to get the NDRectangle back from the current domain object
        rect = current_domain.ndrectangle

        range1 = rect.range(0)
        range2 = rect.range(1)

        self.assertEqual(range1, range_one)
        self.assertEqual(range2, range_two)

        range1 = rect.range("x")
        range2 = rect.range("y")

        # should be the same
        self.assertEqual(range1, range_one)
        self.assertEqual(range2, range_two)

    def test_current_domain_with_ndrectangle_string(self):
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="d1", dtype="S0"),
            tiledb.Dim(name="d2", dtype="S0"),
        )
        ndrect = tiledb.NDRectangle(ctx, dom)

        range_one = ("a", "c")
        range_two = ("b", "db")

        ndrect.set_range(0, range_one[0], range_one[1])
        ndrect.set_range(1, range_two[0], range_two[1])

        self.assertEqual(ndrect.range(0), range_one)
        self.assertEqual(ndrect.range(1), range_two)

        current_domain = tiledb.CurrentDomain(ctx)

        self.assertTrue(current_domain.is_empty)
        current_domain.set_ndrectangle(ndrect)
        self.assertFalse(current_domain.is_empty)

        # let's try to get the NDRectangle back from the current domain object
        rect = current_domain.ndrectangle

        range1 = rect.range(0)
        range2 = rect.range(1)

        self.assertEqual(range1, range_one)
        self.assertEqual(range2, range_two)

        range1 = rect.range("d1")
        range2 = rect.range("d2")

        # should be the same
        self.assertEqual(range1, range_one)
        self.assertEqual(range2, range_two)

    def test_array_schema_with_current_domain_with_ndrectangle(self):
        uri = tempfile.mkdtemp()
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="d", domain=(1, 999), tile=2, dtype=np.int64),
            tiledb.Dim(name="d2", domain=(1, 999), tile=2, dtype=np.int64),
        )
        att = tiledb.Attr(name="a", dtype=np.int64)
        schema = tiledb.ArraySchema(sparse=True, ctx=ctx, domain=dom, attrs=(att,))

        ndrect = tiledb.NDRectangle(ctx, dom)
        range_one = (10, 20)
        range_two = (30, 40)
        ndrect.set_range(0, range_one[0], range_one[1])
        ndrect.set_range(1, range_two[0], range_two[1])

        current_domain = tiledb.CurrentDomain(ctx)
        current_domain.set_ndrectangle(ndrect)
        schema.set_current_domain(current_domain)

        # create the array
        tiledb.Array.create(uri, schema)

        # open the array and check the current domain and the NDRectangle
        A = tiledb.Array(uri, mode="r")

        cd = A.schema.current_domain
        self.assertFalse(cd.is_empty)
        self.assertEqual(cd.type, lt.CurrentDomainType.NDRECTANGLE)

        ndr = cd.ndrectangle
        self.assertEqual(ndr.range(0), range_one)
        self.assertEqual(ndr.range(1), range_two)

        # a 3rd dimension should raise an error
        with self.assertRaises(tiledb.TileDBError):
            ndr.range(2)

        A.close()

    def test_current_domain_evolve(self):
        uri = tempfile.mkdtemp()
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="d", domain=(1, 999), tile=2, dtype=np.int64),
        )

        att = tiledb.Attr(name="a", dtype=np.int64)
        schema = tiledb.ArraySchema(sparse=True, ctx=ctx, domain=dom, attrs=(att,))

        ndrect = tiledb.NDRectangle(ctx, dom)
        range_one = (10, 20)
        ndrect.set_range(0, range_one[0], range_one[1])

        current_domain = tiledb.CurrentDomain(ctx)
        current_domain.set_ndrectangle(ndrect)
        schema.set_current_domain(current_domain)

        tiledb.Array.create(uri, schema)

        new_range = (5, 30)
        new_ndrect = tiledb.NDRectangle(ctx, dom)
        new_ndrect.set_range(0, new_range[0], new_range[1])
        new_current_domain = tiledb.CurrentDomain(ctx)
        new_current_domain.set_ndrectangle(new_ndrect)

        se = tiledb.ArraySchemaEvolution(ctx)
        se.expand_current_domain(new_current_domain)
        se.array_evolve(uri)

        A = tiledb.Array(uri, mode="r")
        s = A.schema
        cd = s.current_domain
        n = cd.ndrectangle
        self.assertEqual(n.range(0), new_range)
        A.close()

    @pytest.mark.skipif(
        tiledb.libtiledb.version()[0] == 2 and tiledb.libtiledb.version()[1] < 27,
        reason="Support for current domain on dense arrays was added in 2.27",
    )
    def test_take_current_domain_into_account_dense_indexing_sc61914(self):
        uri = self.path("test_sc61914")
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="d1", domain=(0, 99), tile=20, dtype=np.int64),
            tiledb.Dim(name="d2", domain=(0, 99), tile=20, dtype=np.int64),
        )
        att = tiledb.Attr(name="a", dtype=np.int64)
        schema = tiledb.ArraySchema(sparse=False, ctx=ctx, domain=dom, attrs=(att,))

        tiledb.Array.create(uri, schema)

        data = np.arange(0, 10000).reshape(100, 100)

        with tiledb.open(uri, "w") as A:
            A[:] = data

        with tiledb.DenseArray(uri, mode="r") as A:
            ndrect = tiledb.NDRectangle(ctx, dom)
            range_one = (10, 20)
            range_two = (30, 35)
            ndrect.set_range(0, range_one[0], range_one[1])
            ndrect.set_range(1, range_two[0], range_two[1])

            current_domain = tiledb.CurrentDomain(ctx)
            current_domain.set_ndrectangle(ndrect)
            A.schema.set_current_domain(current_domain)

            # Define the expected results
            d1_values = range(10, 21)
            d2_values = range(30, 36)
            data = [
                (d1, d2, d1 * 100 + d2)
                for d1, d2 in itertools.product(d1_values, d2_values)
            ]
            expected_df = pd.DataFrame(data, columns=["d1", "d2", "a"])

            expected_array = np.array(
                [
                    [1030, 1031, 1032, 1033, 1034, 1035],
                    [1130, 1131, 1132, 1133, 1134, 1135],
                    [1230, 1231, 1232, 1233, 1234, 1235],
                    [1330, 1331, 1332, 1333, 1334, 1335],
                    [1430, 1431, 1432, 1433, 1434, 1435],
                    [1530, 1531, 1532, 1533, 1534, 1535],
                    [1630, 1631, 1632, 1633, 1634, 1635],
                    [1730, 1731, 1732, 1733, 1734, 1735],
                    [1830, 1831, 1832, 1833, 1834, 1835],
                    [1930, 1931, 1932, 1933, 1934, 1935],
                    [2030, 2031, 2032, 2033, 2034, 2035],
                ]
            )

            assert_array_equal(A, expected_array)
            assert_array_equal(A.df[:, :], expected_df)

            # check indexing the array inside the range of the current domain
            assert_array_equal(A[11:14, 33:35]["a"], expected_array[1:4, 3:5])
            filtered_df = expected_df.query(
                "d1 >= 11 and d1 <= 14 and d2 >= 33 and d2 <= 35"
            ).reset_index(drop=True)
            assert_array_equal(A.df[11:14, 33:35], filtered_df)

            # check only one side of the range
            assert_array_equal(A[11:, :35]["a"], expected_array[1:, :5])
            filtered_df = expected_df.query("d1 >= 11 and d2 <= 35").reset_index(
                drop=True
            )
            assert_array_equal(A.df[11:, :35], filtered_df)

            # check indexing the array outside the range of the current domain - should raise an error
            with self.assertRaises(tiledb.TileDBError):
                A[11:55, 33:34]

            with self.assertRaises(tiledb.TileDBError):
                A.df[11:55, 33:34]

    def test_take_current_domain_into_account_sparse_indexing_sc61914(self):
        uri = self.path("test_sc61914")
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="d1", domain=(0, 99), tile=20, dtype=np.int64),
            tiledb.Dim(name="d2", domain=(0, 99), tile=20, dtype=np.int64),
        )
        att = tiledb.Attr(name="a", dtype=np.int64)
        schema = tiledb.ArraySchema(sparse=True, ctx=ctx, domain=dom, attrs=(att,))

        tiledb.Array.create(uri, schema)

        data = np.arange(0, 2500).reshape(50, 50)

        d1 = np.linspace(10, 59, num=50, dtype=np.int64)
        d2 = np.linspace(10, 69, num=50, dtype=np.int64)
        coords_d1, coords_d2 = np.meshgrid(d1, d2, indexing="ij")

        with tiledb.open(uri, "w") as A:
            A[coords_d1.flatten(), coords_d2.flatten()] = data

        with tiledb.open(uri, mode="r") as A:
            ndrect = tiledb.NDRectangle(ctx, dom)
            range_one = (21, 23)
            range_two = (35, 38)
            ndrect.set_range(0, range_one[0], range_one[1])
            ndrect.set_range(1, range_two[0], range_two[1])

            current_domain = tiledb.CurrentDomain(ctx)
            current_domain.set_ndrectangle(ndrect)
            A.schema.set_current_domain(current_domain)

            expected_array = {
                "d1": np.array([21, 21, 21, 21, 22, 22, 22, 22, 23, 23, 23, 23]),
                "d2": np.array([35, 36, 37, 38, 35, 36, 37, 38, 35, 36, 37, 38]),
                "a": np.array(
                    [571, 572, 573, 574, 621, 622, 623, 624, 671, 672, 673, 674]
                ),
            }

            assert_array_equal(A[:]["d1"], expected_array["d1"])
            assert_array_equal(A[:]["d2"], expected_array["d2"])
            assert_array_equal(A[:]["a"], expected_array["a"])

            expected_df = pd.DataFrame(expected_array)
            assert_array_equal(A.df[:], expected_df)



"""Tests combining fork with tiledb context threads.

Background: the core tiledb library uses threads and it's easy to
experience deadlocks when forking a process that is using tiledb.  The
project doesn't have a solution for this at the moment other than to
avoid using fork(), which is the same recommendation that Python makes.
Python 3.12 warns if you fork() when multiple threads are detected and
Python 3.14 will make it so you never accidentally fork():
multiprocessing will default to "spawn" on Linux.
"""

import subprocess
import sys

import pytest


def run_in_subprocess(code):
    """Runs code in a separate subprocess."""
    script = f"""
import os
import warnings
import tiledb
import multiprocessing

warnings.simplefilter('error')
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

def wrapper_func():
    {code}

wrapper_func()
"""
    result = subprocess.run(
        [sys.executable, "-c", script], capture_output=True, text=True
    )
    if result.returncode != 0:
        print(result.stderr)
    assert result.returncode == 0


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
def test_no_warning_fork_without_ctx():
    """Get no warning if no tiledb context exists."""
    run_in_subprocess(
        """
    pid = os.fork()
    if pid == 0:
        os._exit(0)
    else:
        os.wait()
    """
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
def test_warning_fork_with_ctx():
    """Get a warning if we fork after creating a tiledb context."""
    run_in_subprocess(
        """
    _ = tiledb.Ctx()
    pid = os.fork()
    if pid == 0:
        os._exit(0)
    else:
        os.wait()
    """
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
def test_warning_fork_with_default_ctx():
    """Get a warning if we fork after creating a default context."""
    run_in_subprocess(
        """
    _ = tiledb.default_ctx()
    pid = os.fork()
    if pid == 0:
        os._exit(0)
    else:
        os.wait()
    """
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
def test_no_warning_multiprocessing_without_ctx():
    """Get no warning if no tiledb context exists."""
    run_in_subprocess(
        """
    mp = multiprocessing.get_context("fork")
    p = mp.Process()
    p.start()
    p.join()
    """
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
def test_warning_multiprocessing_with_ctx():
    """Get a warning if we fork after creating a tiledb context."""
    run_in_subprocess(
        """
    _ = tiledb.Ctx()
    mp = multiprocessing.get_context("fork")
    p = mp.Process()
    p.start()
    p.join()
    """
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
def test_warning_multiprocessing_with_default_ctx():
    """Get a warning if we fork after creating a default context."""
    run_in_subprocess(
        """
    _ = tiledb.default_ctx()
    mp = multiprocessing.get_context("fork")
    p = mp.Process()
    p.start()
    p.join()
    """
    )
